/*
 *	XENIX 8086 debugger
 *
 * Modification History:
 *	arw	30 Dec 1981	M000
 *		- writeregs() was loading the addresses of adb's local
 *		  copy of the registers, not the values.  Fixed it.
 *		- Corrected the calls to ptrace in writeregs().
 *	arw	 5 Jan 1982	M001
 *		- Modified delbp(), setbp() and execbkpt() to remove
 *		  special code having to do with odd instruction
 *		  addresses.  Not commented.
 *	arw	 7 May 1982	M002
 *		- Adjusted the notion of `pc' so that it refers to
 *		  the sum of `cs' shifted appropriately and `ip' in
 *		  16 bits.
 *	arw	12 Aug 1982	M003
 *		- Deleted special treatment of SIGINS.  It's for
 *		  PDP-11 floating point only.
 *		- Modified readregs so that floating point registers
 *		  are not read on each breakpoint, but only when
 *		  requested by printfregs/print.c.  This involves
 *		  adding a new function, readfregs.
 *		- Modified execbkpt() to re-read the breakpointed
 *		  instruction after single-stepping it.  This, for
 *		  the system's replacement of interpreted f.p.
 *      JGL     09 Dec 1982     M004
 *              - adb adjusts the user program sp to not show the
 *                stuff stored on the user stack upon breakpoint.
 *                This stuff is moved around if the user moves his
 *                stack pointer.
 *	arw	24 Jan 1983	M005
 *		- Modified readfregs to also read the floating point
 *		  error register.
 */

#include "defs.h"
#include <sys/mmu.h>

MSG		NOFORK;
MSG		ENDPCS;
MSG		BADWAIT;

CHAR		*lp;
INT		sigint;
INT		sigqit;

/* breakpoints */
BKPTR		bkpthead;

REGLIST		reglist[];

CHAR		lastc;
POS		corhdr[];
POS		*endhdr;

INT		fcor;
INT		fsym;
STRING		errflg;
INT		errno;
INT		signo;

L_INT		dot;
STRING		symfil;
INT		wtflag;
INT		pid;
L_INT		expv;
INT		adrflg;
L_INT		loopcnt;

int             oldsp;



/* service routines for sub process control */

getsig(sig)
{	return(expr(0) ? shorten(expv) : sig);
}

INT		userpc=1;
#ifdef USEFPS
INT             fpsim=1;
#endif

runpcs(runmode, execsig)
{
	INT             rc = 0;
	REG BKPTR	bkpt;
	IF adrflg
	THEN userpc = shorten(dot);
	FI
	setbp();
	if (Regchange)
		writeregs();
	printf("%s: running\n", symfil);

	WHILE (loopcnt--)>0
	DO
#ifdef DEBUG
		printf("\ncontinue %d %d\n",userpc,execsig);
#endif
		stty(0,&usrtty);
#ifdef USEFPS
		REP
			fpsim = 0;
			ptrace(runmode,pid,userpc,execsig);
			bpwait();
			IF fpsim
			THEN    rc=0; execsig=signo; userpc = SAMEPC;
			FI
		PER fpsim DONE
		chkerr(); readregs();
#else
		if (userpc == SAMEPC)
			ptrace(runmode,pid,userpc,execsig);
		else
			ptrace(runmode,pid,userpc-((cs-SEGUI)<<4),execsig);
		bpwait(); chkerr(); readregs();
#endif
		/*look for bkpt*/
		IF signo==0 ANDF (bkpt=scanbkpt(pc-1))
		THEN /*stopped at bkpt*/
#ifdef	DEBUG
		     printf ("==== STOPPING....\n");
#endif
		     userpc = pc = bkpt->loc;
		     IF bkpt->flag == BKPTEXEC
			ORF ((bkpt->flag=BKPTEXEC, command(bkpt->comm,':')) ANDF --bkpt->count)
		     THEN execbkpt(bkpt); execsig=0; loopcnt++;
			  userpc = SAMEPC;
		     ELSE bkpt->count = bkpt->initcnt;
			  rc=1;
		     FI
		ELSE rc=0; execsig=signo; userpc = SAMEPC;
		FI
	OD
	return(rc);
}

endpcs()
{
	REG BKPTR	bkptr;
	IF pid
	THEN ptrace(EXIT,pid,0,0); pid=0; userpc=1;
	     FOR bkptr=bkpthead; bkptr; bkptr=bkptr->nxtbkpt
	     DO IF bkptr->flag
		THEN bkptr->flag=BKPTSET;
		FI
	     OD
	FI
}

setup()
{
	close(fsym); fsym = -1;
	IF (pid = fork()) == 0
	THEN
	     ptrace(SETTRC,0,0,0);
	     signal(SIGINT,sigint); signal(SIGQUIT,sigqit);
	     doexec(); exit(0);
	ELIF pid == -1
	THEN error(NOFORK);
	ELSE bpwait(); readregs(); lp[0]=EOR; lp[1]=0;
	     fsym=open(symfil,wtflag);
	     IF errflg
	     THEN printf("%s: cannot execute\n",symfil);
		  endpcs(); error(0);
	     FI
	FI
}

execbkpt(bkptr)
REG BKPTR	bkptr;
{	REG INT		bkptloc;
	REG INT         bkptip;
	unsigned	holder;
#ifdef DEBUG
	printf("exbkpt: %d\n",bkptr->count);
#endif
	bkptloc = bkptr->loc;
	ptrace(WIUSER,pid,bkptloc,bkptr->ins);
	stty(0,&usrtty);
	bkptip = bkptloc - ((cs-SEGUI)<<4);
#ifdef	DEBUG
	printf ("loc = %x cs:ip = %x:%x bkptip=%x\n", bkptloc,
			cs, ip, bkptip);
#endif
	ptrace(SINGLE,pid,bkptip,0);
	bpwait(); chkerr();
	holder = bkptr->ins = ptrace(RIUSER, pid, bkptloc, 0);
	holder = (holder & 0xff00) | BPT;
	ptrace(WIUSER, pid, bkptloc, holder);
	bkptr->flag=BKPTSET;
}


doexec()
{
	STRING		argl[MAXARG];
	CHAR		args[LINSIZ];
	STRING		p, *ap, filnam;
	ap=argl; p=args;
	*ap++=symfil;
	REP	IF rdc()==EOR THEN break; FI
		*ap = p;
		WHILE lastc!=EOR ANDF lastc!=SP ANDF lastc!=TB DO *p++=lastc; readchar(); OD
		*p++=0; filnam = *ap+1;
		IF **ap=='<'
		THEN	close(0);
			IF open(filnam,0)<0
			THEN	printf("%s: cannot open\n",filnam); exit(0);
			FI
		ELIF **ap=='>'
		THEN	close(1);
			IF creat(filnam,0666)<0
			THEN	printf("%s: cannot create\n",filnam); exit(0);
			FI
		ELSE	ap++;
		FI
	PER lastc!=EOR DONE
	*ap++=0;
	execv(symfil, argl);
}

BKPTR	scanbkpt(adr)
{
	REG BKPTR	bkptr;
	FOR bkptr=bkpthead; bkptr; bkptr=bkptr->nxtbkpt
	DO IF bkptr->flag ANDF bkptr->loc==adr
	   THEN break;
	   FI
	OD
	return(bkptr);
}

delbp()
{
	REG INT		a;
	REG BKPTR	bkptr;
	FOR bkptr=bkpthead; bkptr; bkptr=bkptr->nxtbkpt
	DO IF bkptr->flag
	   THEN bkptr->loc;
		ptrace(WIUSER, pid, bkptr->loc, bkptr->ins);
	   FI
	OD
}

setbp()
{
	REG INT		a;
	REG BKPTR	bkptr;
	REG unsigned	holder;

	FOR bkptr=bkpthead; bkptr; bkptr=bkptr->nxtbkpt
	DO IF bkptr->flag
	   THEN a = bkptr->loc;
		holder = bkptr->ins = ptrace(RIUSER, pid, a, 0);
		holder = (holder & 0xff00) | BPT;
		ptrace(WIUSER, pid, a, holder);
		IF errno
		THEN prints("cannot set breakpoint: ");
		     printf("err # %d  ", errno);
		     perror("");
		     psymoff(leng(bkptr->loc),ISYM,"\n");
		FI
	   FI
	OD
}

bpwait()
{
	REG INT w;
	INT stat;

	signal(SIGINT, 1);
	WHILE (w = wait(&stat))!=pid ANDF w != -1 DONE
	signal(SIGINT,sigint);
	gtty(0,&usrtty);
	stty(0,&adbtty);
	IF w == -1
	THEN pid=0;
	     errflg=BADWAIT;
	ELIF (stat & 0177) != 0177
	THEN IF signo = stat&0177  /* can not be restarted */
	     THEN sigprint();
	     FI
	     IF stat&0200
	     THEN prints(" - core dumped");
		  close(fcor);
		  setcor();
	     FI
	     pid=0;
	     errflg=ENDPCS;
	ELSE signo = stat>>8;      /* can be restarted */
#ifdef USEFPS
	     IF signo == SIGINS
	     THEN
		 /* may be a fp instruction */
		IF fpsim=hasfps() THEN return; FI
	     FI
#endif
#ifdef DEBUG
	     prints (" ** Have 177 match **");
#endif
	     IF signo!=SIGTRC
	     THEN sigprint();
	     ELSE signo=0;
	     FI
	     flushbuf();
	FI
}

static INT fpsaved;

readregs()
{
	/*get register values from pcs*/
	struct user *unull = 0;
	REG i;
	INT axptr;
		/* read pointer to regs in u_ area */
	axptr = ptrace (RUREGS, pid, &(unull->u_aAX), 0);
	axptr &= OFFMASK;
		/* read user regs in u_ area */

	ax = ptrace (RUREGS, pid, axptr + 2*RAX, 0);
	bx = ptrace (RUREGS, pid, axptr + 2*RBX, 0);
	cx = ptrace (RUREGS, pid, axptr + 2*RCX, 0);
	di = ptrace (RUREGS, pid, axptr + 2*RDI, 0);
	si = ptrace (RUREGS, pid, axptr + 2*RSI, 0);
	ds = ptrace (RUREGS, pid, axptr + 2*RDS, 0);
	ss = ptrace (RUREGS, pid, axptr + 2*RSS, 0);
	sp = ptrace (RUREGS, pid, axptr + 2*RSP, 0);
		/* now, lets read the regs on the user stack */
	dx = ptrace (RDUSER, pid, sp + URDX, 0);
	es = ptrace (RDUSER, pid, sp + URES, 0);
	bp = ptrace (RDUSER, pid, sp + URBP, 0);
	cs = ptrace (RDUSER, pid, sp + URCS, 0);
	ip = ptrace (RDUSER, pid, sp + URIP, 0);		/* M002 */
	pc = ((cs-SEGUI) << 4) + ip;				/* M002 */
	flag = ptrace (RDUSER, pid, sp + URFLAG, 0);
	oldsp = sp;                                             /* M004 */
	sp += URBC;                                             /* M004 */
	Regchange = 0;
	fpsaved = 0;						/* M003 */
}

#ifdef	HAVEFP
readfregs()							/* M003 */
{
	REG	i, *j;						/* M003 */

	i = ((char *) &corhdr[0].u_fper - (char *) &corhdr[0]);	/* M005 */
	corhdr[0].u_fper = ptrace(RUREGS,pid,i,0);		/* M005 */

	/* read floating point status	*/
	j = &corhdr[0].u_fps;					/* M003 */
	FOR i=FROFF; i<FRLEN+FROFF; i += sizeof (POS)
	DO	*j++ = ptrace(RUREGS,pid,i,0);			/* M003 */
	OD
	fpsaved = 1;						/* M003 */
}
#endif



/**     writeregs - write register values back into user process
 *
 *      writeregs is called just before firing up the user process,
 *      iff the register values have changed since they were last
 *      read up.  Note that this means writeregs is not called for ":r".
 *
 *      returns 0 if OK
 *              1 if stack would be invalid
 */

writeregs()
{
	struct user *unull = 0;
	REG	i, *j;						/* M003 */
	INT	axptr;
		/* read pointer to regs in u_ area */
#ifdef	DEBUG
	printf("IN WRITEREGS\n");
#endif
	axptr = ptrace (RUREGS, pid, &(unull->u_aAX), 0);
	axptr &= OFFMASK;
		/* write user regs from u_ area */
	ptrace (WUREGS, pid, axptr + 2*RAX, ax);
	ptrace (WUREGS, pid, axptr + 2*RBX, bx);
	ptrace (WUREGS, pid, axptr + 2*RCX, cx);
	ptrace (WUREGS, pid, axptr + 2*RDI, di);
	ptrace (WUREGS, pid, axptr + 2*RSI, si);
	ptrace (WUREGS, pid, axptr + 2*RDS, ds);
	ptrace (WUREGS, pid, axptr + 2*RSS, ss);
	sp -= URBC;                                             /* M004 */
	if (sp != oldsp) {
		/* had a stack pointer change */
#ifdef	DEBUG
		printf ("stack from %x to %x\n", oldsp, sp);
#endif
		if (ptrace(RDUSER, pid, sp + URFLAG, 0))
			return(1);                      /* invalid stack */
		ptrace (WUREGS, pid, axptr + 2*RSP, sp);
	}
		/* now, lets write the regs on the user stack */
	ptrace (WDUSER, pid, sp + URDX, dx);
	ptrace (WDUSER, pid, sp + URES, es);
	ptrace (WDUSER, pid, sp + URBP, bp);
	ptrace (WDUSER, pid, sp + URCS, cs);
	ptrace (WDUSER, pid, sp + URFLAG, flag);

	/* floating point registers can't be modified currently.*/
#ifdef	HAVEFP
/*      j = &corhdr[0].u_fps;
/*      IF fpsaved THEN
/*		FOR i=FROFF; i<FRLEN+FROFF; i += sizeof (POS)
/*              DO ptrace(WUREGS, pid, i, *j++); OD
/*      FI      */
#endif
	return(0);
}

#ifdef USEFPS
#define USIGOFF	((struct user *)0->u_signal[SIGINS])
hasfps() /* user program has floating point simulator */
{
	static CHAR once=1;
	static CHAR hasit=0;
	INT i;
	IF once
	THEN
		i = ptrace(RUREGS,pid,USIGOFF,0);
		IF i==0 ORF i==1
		THEN hasit=0;
		ELSE hasit=1;
		FI
		once=0;
	FI
	return(hasit);
}
#endif
