/*
 *
 *	XENIX debugger
 *
 * Copyright (C) Microsoft Corporation, 1982
 */

/*
 *  JGL and Hans	19,20 Dec, 1982
 *		- Major rework for x.out format.  no M000s
 *
 *	arw		 3 Feb 1983	M000
 *		- Fixed cntsym to return error when problems are
 *		  encountered while reading the symbol table.
 *	Hans		 1 Mar 1983	M001
 *		- translated some stuff to C.
 *		- hacked around symbol matching to deal with segments
 *	Hans		10 Mar 1983	M002
 *		- more translation
 *		- made psymoff and valpr use IDISP.
 *	Hans		22 Mar 1983	M003
 *		- fixed freadsym to fix a sign extension bug
 *		- made it do appropriate stuff for SEGSYM
 */

#include "defs.h"
#include <sys/relsym.h>

MSG		BADFIL;

SYMTAB		symbol;
BOOL		localok;
INT		lastframe;
SYMSLAVE		*symvec;
POS		maxoff;
L_INT		maxstor;

/* symbol management */
L_INT		symbas;
L_INT           symcnt;         /* count of syms left during sequential scan */
L_INT		symnum;
L_INT		localval;
char            symrqd = -1;
SYMTAB		symbuf[SYMSIZ];
SYMPTR		symnxt;
SYMPTR		symend;


INT		fsym;
FILE            *symstream;
STRING		errflg;
POS		findsym();
CHAR		*IDISP;			/*M002*/


/* symbol table and file handling service routines */

longseek(f, a)
L_INT	a;
{
	return(lseek(f, a, 0) != -1);
}

valpr(v,idsp)
SYMV	v;
{
	POS		d;
	d = findsym(v,idsp);
	if(d < maxoff)
	{
		printf("%.8s", symbol.symc);
		if(d)
		{			/*M002*/
			printf("+");
			printf(IDISP, d);
		}
	}
}


/***    localsym
 *
 *      something to do with local symbols...
 */

localsym(cframe)
L_INT cframe;
{
	INT symflg;
	WHILE nextsym() ANDF localok
		ANDF symbol.symc[0]!='~'
		ANDF (symflg=symbol.symf)!=037
	DO IF symflg>=2 ANDF symflg<=4
	   THEN localval=symbol.symv;
		return(TRUE);
	   ELIF symflg==1
	   THEN localval=leng(shorten(cframe)+symbol.symv);
		return(TRUE);
	   ELIF symflg==20 ANDF lastframe
	   THEN localval=leng(lastframe+2*symbol.symv-10);
		return(TRUE);
	   FI
	OD
	return(FALSE);
}



/* *** psymoff - print a symbol and offset.
 *
 *	given a value, a type and a trailing string, find the
 *	nearest symbol using 'findsym'.  print the symbol name,
 *	the offset, and the trailing string.
 */
psymoff(v,type,s)
L_INT v; int type; char *s;
{
	POS		w;
	w = findsym(v,type);
	if(w >= maxoff)
		printf(LPRMODE,v);
	else
	{
		printf("%.8s", symbol.symc);
		if(w)			/*M002*/
		{
			printf("+");
			printf(IDISP,w);
		}
	}
	printf(s);
}



/***    findsym - find the nearest symbol
 *
 *      findsym finds the symbol of type "type" nearest to and <= to
 *      "svalue".  It returns with the entry in "symbol".  The return
 *      code is the difference between "svalue" and the found value.
 */

POS	findsym(svalue,type)
    SYMV     svalue;
    INT     type;
{
	long		diff, value, symval;
	INT		symtyp;
	REG SYMSLAVE	*symptr;
	SYMSLAVE	*symsav;
	register short cnt;
	static long lval;

	value=svalue; diff = 0377777L; symsav=0;
	if(type!=NSYM && (symptr=symvec))
	{
		while(diff && (symtyp=symptr->typslave)!=ESYM)
		{
			if(symtyp==type)
			{
				/* hacked around to make sure same
				   segment M001 */
				symval=symptr->valslave;
				if(
#ifdef SEGSYM
				   (value&0xffff0000L)==(symval&0xffff0000) &&
#endif
				   value-symval<diff &&
				   value>=symval)
				{
					diff = value-symval;
					symsav=symptr;
				}
			}
			symptr++;
		}
		if(symsav)
		{
			value = (((long)symsav->seekhi)<<16) |
				(long)(symsav->seeklo);
			fseek(symstream, value, 0);
			freadsym(&symbol);
		}
	}
	return(shorten(diff));
}




nextsym()
{
	if((--symcnt)<0)
		return(FALSE);
	else
		return(fseek(symstream, symbas+(symnum-symcnt)*SYMTABSIZ)!=1 &&
			fread(&symbol, SYMTABSIZ, 1, symstream)==1);
}



/* sequential search through file */

long symoff;

/***    symset - "rewind" symbol table list
 *
 *      symset is called as a preamble to calling 'symget'
 */

symset()
{
	symcnt = symnum;
	symoff = 0;
	fseek(symstream, symbas, 0);
}



/***    symget - get index of next symbol
 *
 *      symptr = symget(offadr)
 *
 *      symget makes sure that the next symbol in somewhere in symbuf[]
 *      and returns its index.
 *      If offadr is non-null the offset, from the beginning of the symtab,
 *      is stored at *offadr;
 *
 *      returns  0 if no more symbols
 *               pointer to symbol in symbuf if more exist
 */

SYMPTR  symget(offadr)
    long *offadr;
{
	static struct symtab symb;
	static long lval;
	register short cnt;

	if (offadr)
		*offadr = symoff;
	if (--symcnt < 0)
		return(0);
	symoff += freadsym(&symb);
	return(&symb);
}


/*
 *  *** freadsym  fread a symbol into a symtab struct
 *	read one symbol from the symbol table in x.out format
 *	and turn it into internal format.
 *	return the total size read.
 */
freadsym(symb)
struct symtab *symb;
{
	static struct sym xoutsym;
	register char *cp;
	register int count;
	/* note that args are backwards so result is byte count, not
	   object count */
	count = fread (&xoutsym, 1, sizeof(xoutsym), symstream);
	symb->symf = xoutsym.s_type;
/*M003*/
#ifdef SEGSYM
	symb->symv = xoutsym.s_value;
#else
	symb->symv = ((unsigned short)xoutsym.s_value) + ((xoutsym.s_value>>12)&0xffff0L);
#endif
	cp = symb->symc;
	while (cp< &(symb->symc[8]) && (*cp++ = getc(symstream)))
		count++;
	if (cp[-1])
		while (getc(symstream))
			count++;
	return (count+1);  /* count the null */
}

/***    cntsyms - count number of symbols in exe file
 *
 *      since the symbol records are of variable length we need
 *      to read them all to find out how many there are.  Keep this
 *      value around, its expensive to compute!
 *
 *      cntsyms(size)
 *
 *              size = length, in bytes, of symbol table area
 *
 *      returns         0 if file format is screwed up
 *                      cnt = symbol count
 */

cntsyms(size)
    long size;
{
	register int i;
	register int c;
	register int symcnt;
	static struct symtab symb;

	fseek(symstream, symbas, 0);            /* see to the right spot */
	symcnt = 0;
	while (size > 0) {
		size -= freadsym(&symb);
		if (ferror(symstream) || feof(symstream))	/* M000 */
			return(0);				/* M000 */
		symcnt++;
	}
	if (size == 0)
		return(symcnt);
	return(0);
}



/**     symread - fill symbuf[]
 *
 *      symread is called by symget when the buffer symbuf[] is empty.
 */

symread()
{
	INT		symlen;

	IF (symlen=read(fsym,symbuf,sizeof symbuf))>=SYMTABSIZ
	THEN	symnxt = symbuf;
		symend = &symbuf[symlen/SYMTABSIZ];
		return(TRUE);
	ELSE	return(FALSE);
	FI
}
